import os, sys, math, random, time, pygame
#import binascii  #for debugging
from pygame.locals import *
from pygame.compat import geterror

import const, util, image, var, explosion, weapon

class Item(pygame.sprite.DirtySprite):    #generic item, both spiders and bricks
    def __init__(self, max_hits):
        pygame.sprite.DirtySprite.__init__(self)
        self.image = None
        self.rect = None
        self.hits = 0
        self.max_hits = max_hits    #max hits the item takes before getting killed
        self.collided = False
        self.enemy = False
        self.killer = False    #this item kills you
        self.shrinker = False  #this item shrinks the paddle (permanently until losing life)
        self.dead = False

    def hit(self):
        self.hits += 1
        if self.max_hits == 0:
            return False
        return self.hits >= self.max_hits

    def fix_collides(self, collides):
        if collides:
            if self.collided:   #ignore duplicate collisions (workaround)
                return False
            self.collided = True
            return True
        else:
            self.collided = False
            return False

    def destroy(self):
        if self.dead:
            return
        self.dead = True
        var.game.bricks.remove(self)
        var.game.allsprites.remove(self)
        
    def die(self):
        e = explosion.Explosion(self.rect, False, False)
        if self.enemy:
            var.game.enemies -= 1
        self.destroy()    

class Brick(Item):
    def __init__(self, max_hits):
        Item.__init__(self, max_hits)        
        self.layer = 0

    def init(self, x, y):
        self.rect = self.image.get_rect()
        self.rect[0] = x
        self.rect[1] = y

    def collides(self, sprite):
        return self.fix_collides(pygame.sprite.collide_rect(self, sprite))

class UnbreakableBrick(Brick):
    def __init__(self, x, y):    
        Brick.__init__(self, 0)
        self.image = image.images.unbreakable_brick
        self.init(x, y)

class BreakableBrick(Brick):
    def __init__(self, x, y, max_hits):
        Brick.__init__(self, max_hits)
        self.images = image.images.breakable_brick
        self.image = self.images[0]
        self.init(x, y)
       
    def hit(self):
        self.hits += 1
        if self.hits < len(self.images):
            self.image = self.images[self.hits]
            self.dirty = 1
        return self.hits >= self.max_hits

class MovingBrick(Brick):
    def __init__(self, x, y, image, max_hits, vel):    
        Brick.__init__(self, max_hits)
        self.image = image
        self.init(x, y)
        self.moving = False
        self.delay = 0
        self.vel = vel
        self.shrinker = True
        
    def hit(self):
        self.hits += 1
        if not self.moving:
            self.moving = True
        return self.hits >= self.max_hits

    def update(self):
        if not self.moving:
            return
        self.delay += 1
        if self.delay % 2 == 0:
            self.dirty = 1
            self.rect[0] += self.vel[0]
            self.rect[1] += self.vel[1]

class FallingBrick(MovingBrick):
    def __init__(self, x, y, max_hits):    
        MovingBrick.__init__(self, x, y, image.images.falling_brick, max_hits, [0, 1])
    
    def update(self):
        MovingBrick.update(self)
        if self.moving and self.rect[1] >= const.SCREEN_HEIGHT:
            self.destroy()

class BubbleBrick(MovingBrick):
    def __init__(self, x, y, max_hits):    
        MovingBrick.__init__(self, x, y, image.images.bubble_brick, max_hits, [0, 0])
        while self.vel[0] == 0 and self.vel[1] == 0:
            self.vel = [random.randrange(-2, 2), random.randrange(-2, 2)]
    
    def update(self):
        MovingBrick.update(self)
        if not self.moving:
            return
        if self.rect[0] <= 0 or self.rect[0]+self.rect[2] >= const.SCREEN_WIDTH:
            self.vel[0] = -self.vel[0]
            self.rect[0] += self.vel[0]
        if self.rect[1] <= const.SCREEN_Y0 or self.rect[1]+self.rect[3] >= const.SCREEN_HEIGHT:
            self.vel[1] = -self.vel[1]
            self.rect[1] += self.vel[1]

class DynamiteBrick(Brick):
    def __init__(self, x, y):    
        Brick.__init__(self, 1)
        self.image = image.images.dynamite_brick
        self.init(x, y)

    def hit(self):
        e = explosion.Explosion(self.rect, False, True)
        for brick in list(var.game.bricks):
            if pygame.sprite.collide_circle_ratio(5.0)(self, brick):
                brick.die()
                var.game.update_score(1)
        #self.destroy()  #it's destroyed elsewhere?

class MutableBrick(Brick):
    def __init__(self, x, y):    
        Brick.__init__(self, 1)
        self.image = image.images.mutable_brick
        self.init(x, y)
        
    def hit(self):
        self.hits += 1
        return True

    def die(self):
        Item.die(self)
        r = random.randrange(0, 14)
        if r < 10:
            t = 'z'
        else:
            if var.game.level < 9:  #spiders w and v were not introduced yet
                types = {10:"y", 11:"x", 12:"y", 13:"x"}
            else:
                types = {10:"y", 11:"x", 12:"w", 13:"v"}
            t = types[r]
        spider = Spider(self.rect[0], self.rect[1], 0, t)
        var.game.enemies += 1
        var.game.bricks.append(spider)
        var.game.allsprites.add(spider, layer=spider.layer)

class Spider(Item):
    def __init__(self, x, y, n, type):
        Item.__init__(self, 2)
        self.images = [ image.images.spiders[type].subsurface(40 + i*128, 70 + (n%8)*128, 45, 45) for i in range(12) ]
        self.type = type #z=regular, x=green (very fast), y=red (fast, tough), v=blue (slow, tough), u=purple (mutable), w=yellow (shooting)
        if type == 'z' or type == 'w' or type == 'u':
            self.vel_minmax = [0, 3]   #min and max velocity (abs! because it can be with minus)
        elif type == 'y':
            self.max_hits = 5
            self.vel_minmax = [3, 6]
        elif type == 'x':
            self.vel_minmax = [6, 10]
        elif type == 'v':
            self.max_hits = 10
            self.vel_minmax = [1, 1]

        dirs = ([-1,-1],[-1,0],[-1,1],[0,-1],[0,1],[1,1],[1,0],[1,-1])   #all directions except (0,0)
        d = random.randrange(0, len(dirs))
        self.vel = [random.randint(self.vel_minmax[0], self.vel_minmax[1])*dirs[d][0], random.randint(self.vel_minmax[0], self.vel_minmax[1])*dirs[d][1]]  #choose random positive values and multiply by direction
        self.image = self.images[0]
        self.rect = self.image.get_rect()
        self.rect[0] = x
        self.rect[1] = y
        self.radius = 15
        self.frame = 0
        self.delay = 0
        self.max_delay = random.randrange(12, 22)
        self.enemy = True
        self.layer = 1
        self.killer = True
        self.voff = ((1,0), (-1,0), (0,1), (0,-1))    #random velocity change (only 1 direction)
        #self.id = binascii.hexlify(os.urandom(16))   #for debugging
        
    def update(self):
        if var.game.bonusdata["freeze"]["cnt"] > 0:   #do not move if frozen (by a bonus)
            return
        self.dirty = 1
        self.delay += 1
        if self.delay > self.max_delay:
            self.delay = 0
            self.frame += 1
            if self.frame >= 8:
                self.frame = 0
            self.image = self.images[self.frame]

            if self.hits == 0:
                return
            a = random.randrange(0, 10)  #a chance of random change of direction
            if a < len(self.voff): 
                if self.type != 'v':   #not a slow spider
                    if abs(self.vel[0] + self.voff[a][0]) >= self.vel_minmax[0] and abs(self.vel[0] + self.voff[a][0]) <= self.vel_minmax[1]:
                        self.vel[0] += self.voff[a][0]
                    if abs(self.vel[1] + self.voff[a][1]) >= self.vel_minmax[0] and abs(self.vel[1] + self.voff[a][1]) <= self.vel_minmax[1]:
                        self.vel[1] += self.voff[a][1]
                else:
                    self.vel[0], self.vel[1] = self.voff[a][0], self.voff[a][1]   #slow spiders can only change velocity, not increase
                
            #if the spider bumps into the edge of the screen, go back        
            #simple self.vel[i]=-self.vel[i] doesn't work in some cases, explicit if/else is more reliable
            if self.rect[0] <= 0:
                if self.vel[0] == 0:
                    self.vel[0] = 1
                else:    
                    self.vel[0] = abs(self.vel[0])
            elif self.rect[0]+self.rect[2] >= const.SCREEN_WIDTH:
                if self.vel[0] == 0:
                    self.vel[0] = -1
                else:    
                    self.vel[0] = -abs(self.vel[0])
            if self.rect[1] <= const.SCREEN_Y0:
                if self.vel[1] == 0:
                    self.vel[1] = 1
                else:    
                    self.vel[1] = abs(self.vel[1])
            elif self.rect[1]+self.rect[3] >= const.SCREEN_HEIGHT:
                if self.vel[1] == 0:
                    self.vel[1] = -1
                else:    
                    self.vel[1] = -abs(self.vel[1])

            self.rect[0] += self.vel[0]    
            self.rect[1] += self.vel[1]
            #if self.vel[0] != 0 or self.vel[1] != 0:
            #    print (str(self.id) + " rect=" + str(self.rect) + " vel=" + str(self.vel))

    def collides(self, sprite):
        return self.fix_collides(pygame.sprite.collide_circle(self, sprite))

class ShootingSpider(Spider):
    def __init__(self, x, y, n, type):
        Spider.__init__(self, x, y, n, type)
        self.cnt = 0
        self.paused = False

    def pause(self, mode):
        if self.hits > 0:
            self.paused = mode

    def update(self):
        Spider.update(self)
        if self.hits == 0 or self.paused:
            return
        self.cnt += 1
        if self.cnt > 100:
            self.cnt = 0
            m = weapon.SpiderMissile(self.rect)
            var.game.allsprites.add(m)

class MutableSpider(Spider):
    def __init__(self, x, y, n, type):
        Spider.__init__(self, x, y, n, type)

    def die(self):
        Item.die(self)
        var.game.bricks_todo.append(self.rect)

    
